import numpy as np
import cv2
from bacs import bacs
from ...helpers import eliminate_2pi
from ...world.world import World
from ...world.rotation import Rotation


class BundleAdjustment:

    def load(self, world: World, sigma_x: float = 1.0):
        self.N = len(world.links)
        self.l = np.zeros((3, self.N), dtype=np.float32)
        self.ict = np.zeros((self.N, 3), dtype=int)
        self.macs = list(set(link.mac for link in world.links))
        self.point_ids = list(set(link.id for link in world.links))
        for n, link in enumerate(world.links):
            calibration = world.cameras[link.mac].calibration
            K = np.array(calibration.intrinsics.matrix)
            D = np.array(calibration.intrinsics.distortion)
            self.ict[n, :] = self.point_ids.index(link.id), 0, self.macs.index(link.mac)
            point = np.array(link.image_point.tuple)
            point = cv2.undistortPoints(point.reshape(-1, 1, 2), K, D)
            self.l[:, n] = cv2.convertPointsToHomogeneous(point.reshape(-1, 2))
        self.Sll = np.array([[[sigma_x**2, 0, 0], [0, sigma_x**2, 0], [0, 0, 0]] for _ in range(self.N)])
        self.I = len(self.point_ids)
        self.Xa = np.zeros((4, self.I))
        for link in world.links:
            self.Xa[:, self.point_ids.index(link.id)] = (link.world_point.x, link.world_point.y, link.world_point.z, 1)
        self.Ma = [
            np.vstack((
                np.hstack((
                    world.cameras[mac].calibration.rotation.R,
                    np.array([world.cameras[mac].calibration.extrinsics.translation]).T,
                )),
                [0, 0, 0, 1],
            ))
            for mac in self.macs
        ]
        self.P = [np.eye(3, 4)]
        self.sigma_h = (0, 0, 0, 0, 0, 0, 0) if any(c.locked for c in world.cameras.values()) else (1, 1, 1, 1, 1, 1, 1)
        self.sigma_c = [s
                        for camera in world.cameras.values()
                        for s in ([0] * 6 if camera.locked else [np.deg2rad(1), np.deg2rad(1), None, None, None, 0.1])
                        ]

    def adjust(self, max_iterations: int = 10):
        _, self.Xa, self.Ma, _, s0dsq, *_ = bacs(
            self.l, self.ict, self.Sll, self.Xa, self.Ma, self.P,
            max_iterations=max_iterations,
            sigma_h=self.sigma_h,
            sigma_c=self.sigma_c)
        return np.sqrt(s0dsq) if s0dsq is not None else None

    def dump(self, world: World) -> tuple[float, float]:
        for link, (i, *_) in zip(world.links, self.ict):
            link.world_point.x = self.Xa[0, i] / self.Xa[-1, i]
            link.world_point.y = self.Xa[1, i] / self.Xa[-1, i]
            link.world_point.z = self.Xa[2, i] / self.Xa[-1, i]

        max_dist = 0
        max_angle = 0
        for mac, m in zip(self.macs, self.Ma):
            calibration = world.cameras[mac].calibration
            old_omega, old_phi, old_kappa = (calibration.rotation * calibration.intrinsics.rotation.T).euler
            new_total_rotation = Rotation(R=m[:3, :3].tolist())
            new_omega, new_phi, new_kappa = (new_total_rotation * calibration.intrinsics.rotation.T).euler
            max_dist = max(max_dist, abs(calibration.extrinsics.translation[0] - m[0, -1]))
            max_dist = max(max_dist, abs(calibration.extrinsics.translation[1] - m[1, -1]))
            max_dist = max(max_dist, abs(calibration.extrinsics.translation[2] - m[2, -1]))
            max_angle = max(max_angle, abs(eliminate_2pi(new_omega - old_omega)))
            max_angle = max(max_angle, abs(eliminate_2pi(new_phi - old_phi)))
            max_angle = max(max_angle, abs(eliminate_2pi(new_kappa - old_kappa)))
            calibration.extrinsics.tilt = Rotation.from_euler(omega=new_omega, phi=new_phi, kappa=0)
            calibration.extrinsics.yaw = new_kappa
            calibration.extrinsics.translation[:] = m[:3, -1]

        return max_dist, max_angle
