import numpy as np
from .frame import Frame
from .point import Point
from .solve import solve


def adjust(frames_approx, points_approx, observations, lock=[]):

    frame_ids = np.unique([frame_id for frame_id, _, _ in observations])
    point_ids = np.unique([point_id for _, point_id, _ in observations])

    F = len(frame_ids)
    U = 3 * F + 2 * len(point_ids)
    N = 2 * len(observations)

    l = np.zeros((N, 1))
    l0 = np.zeros((N, 1))
    A = np.zeros((N, U))
    x0 = np.zeros((U, 1))
    p = np.ones(N)

    for frame_idx, frame_id in enumerate(frame_ids):

        frame0 = frames_approx[frame_id]
        x0[3 * frame_idx + 0, 0] = frame0.x
        x0[3 * frame_idx + 1, 0] = frame0.y
        x0[3 * frame_idx + 2, 0] = frame0.yaw

    for point_idx, point_id in enumerate(point_ids):

        point0 = points_approx[point_id]
        x0[3 * F + 2 * point_idx + 0, 0] = point0.x
        x0[3 * F + 2 * point_idx + 1, 0] = point0.y

    for u, (frame_id, point_id, point_) in enumerate(observations):

        l[2 * u + 0, 0] = point_.x
        l[2 * u + 1, 0] = point_.y

        point0 = points_approx[point_id]
        frame0 = frames_approx[frame_id]

        point0_ = point0.transform(frame0.invert().M)
        l0[2 * u + 0, 0] = point0_.x
        l0[2 * u + 1, 0] = point0_.y

        dx = point0.x - frame0.x
        dy = point0.y - frame0.y
        sin = np.sin(frame0.yaw)
        cos = np.cos(frame0.yaw)

        # [tx, ty, phi] ...
        frame_idx = np.where(frame_ids == frame_id)[0][0]
        A[2 * u + 0, 3 * frame_idx + 0] = -cos
        A[2 * u + 0, 3 * frame_idx + 1] = -sin
        A[2 * u + 0, 3 * frame_idx + 2] = -dx * sin + dy * cos
        A[2 * u + 1, 3 * frame_idx + 0] = sin
        A[2 * u + 1, 3 * frame_idx + 1] = -cos
        A[2 * u + 1, 3 * frame_idx + 2] = -dx * cos - dy * sin

        # ... [x, y] ...
        point_idx = np.where(point_ids == point_id)[0][0]
        A[2 * u + 0, 3 * F + 2 * point_idx + 0] = cos
        A[2 * u + 0, 3 * F + 2 * point_idx + 1] = sin
        A[2 * u + 1, 3 * F + 2 * point_idx + 0] = -sin
        A[2 * u + 1, 3 * F + 2 * point_idx + 1] = cos

    useful_locks = set(lock).intersection(frame_ids)
    if not any(useful_locks):
        A = np.vstack((A, np.eye(3 * F, U)))
        p = np.hstack((p, 0.001 * np.ones(3 * F)))
        l = np.vstack((l, x0[:3 * F]))
        l0 = np.vstack((l0, x0[:3 * F]))
    else:
        for frame_id in useful_locks:
            frame_idx = np.where(frame_ids == frame_id)[0][0]
            A = np.vstack((A, np.zeros((3, U))))
            A[-3:, 3 * frame_idx: 3 * frame_idx + 3] = np.eye(3)
            p = np.hstack((p, 1000 * np.ones(3)))
            l = np.vstack((l, x0[3 * frame_idx: 3 * frame_idx + 3]))
            l0 = np.vstack((l0, x0[3 * frame_idx: 3 * frame_idx + 3]))

    x = x0 + solve(A, p, l - l0)

    for frame_idx, frame_id in enumerate(frame_ids):
        frames_approx[frame_id] = Frame.from_components(x[3 * frame_idx + 0, 0],
                                                        x[3 * frame_idx + 1, 0],
                                                        x[3 * frame_idx + 2, 0])
    for point_idx, point_id in enumerate(point_ids):
        points_approx[point_id] = Point.from_components(x[3 * F + 2 * point_idx + 0, 0],
                                                        x[3 * F + 2 * point_idx + 1, 0])

    sigma_0_estimated = np.linalg.norm(l - l0) / np.sqrt(N - U)

    return frames_approx, points_approx, sigma_0_estimated
