from ...helpers import eliminate_2pi
from ...world.world import World
from .helmert.adjustment import adjust
from .helmert.approximation import find_approximation
from .helmert.frame import Frame
from .helmert.point import Point as HelmertPoint


class HelmertAdjustment:

    def load(self, world: World):
        self.frames = {}
        self.points = {}
        self.observations = []
        self.locks = []
        for mac, camera in world.cameras.items():
            if camera.calibration is None or not camera.calibration.is_complete:
                continue
            self.frames[mac] = Frame.from_components(
                camera.calibration.extrinsics.translation[0],
                camera.calibration.extrinsics.translation[1],
                camera.calibration.extrinsics.yaw,
            )
            if camera.locked:
                self.locks.append(mac)
        for link in world.links:
            self.points[link.id] = HelmertPoint.from_components(x=link.world_point.x, y=link.world_point.y)
            local = self.points[link.id].transform(self.frames[link.mac].invert().M)
            self.observations += [(link.mac, link.id, local)]

    def find_approximation(self):
        lock = self.locks[0] if any(self.locks) else None
        self.frames = find_approximation(self.observations, lock=lock)
        for mac, link_id, local in self.observations:
            self.points[link_id] = local.transform(self.frames[mac].M)

    def adjust(self, iterations: int = 1) -> float:
        for _ in range(iterations):
            self.frames, self.points, sigma_0_estimated = adjust(
                self.frames, self.points, self.observations, lock=self.locks)
        return sigma_0_estimated

    def dump(self, world: World) -> tuple[float, float]:
        for link in world.links:
            link.world_point.x = self.points[link.id].x
            link.world_point.y = self.points[link.id].y

        max_dist = 0
        max_angle = 0
        for mac, camera in world.cameras.items():
            max_dist = max(max_dist, abs(camera.calibration.extrinsics.translation[0] - self.frames[mac].x))
            max_dist = max(max_dist, abs(camera.calibration.extrinsics.translation[1] - self.frames[mac].y))
            max_angle = max(max_angle, abs(eliminate_2pi(camera.calibration.extrinsics.yaw - self.frames[mac].yaw)))
            if mac in self.frames:
                camera.calibration.extrinsics.translation[0] = self.frames[mac].x
                camera.calibration.extrinsics.translation[1] = self.frames[mac].y
                camera.calibration.extrinsics.yaw = self.frames[mac].yaw

        return max_dist, max_angle
