import numpy as np
import asyncio
import pycurl
import io
import json
import uuid
import requests
from .actor import Actor
from ..world.world import World
from ..world.camera import Camera, CameraSynchronization
from ..world.image import Image, Size
from ..world.rotation import Rotation
from ..world.calibration import Calibration, Intrinsics


class CameraDownloader(Actor):

    interval: float = 0.01

    async def step(self, world: World):

        if any(world.download_queue):

            mac = world.download_queue.pop(0)
            camera = world.cameras[mac]

            if camera.synchronization is None or camera.synchronization.time < world.time - 10.0:
                loop = asyncio.get_running_loop()
                camera.synchronization = await loop.run_in_executor(None, lambda: self.synchronize(world, camera))
                if camera.synchronization is None:
                    return

            timeout = 1
            url = f'http://{camera.network.ip}:8080/image?maxtime={timeout}&killtime=2.5'
            loop = asyncio.get_event_loop()
            try:
                header, content = await loop.run_in_executor(None, self.get, url, timeout)
            except pycurl.error:
                self.log.warning(f'image download error from {url}')
                return
            if header['mac'].lower() != mac.lower():
                self.log.warning(f'image download error from {url} (MAC mismatch)')
                return

            intrinsics = json.loads(header['K'])
            imu_calibration = [CameraDownloader.parse_imu(header[f'imu{i}']) for i in range(6)]
            tilt = CameraDownloader.compute_tilt(imu_calibration, CameraDownloader.parse_imu(header['imu']))
            size = Size(
                width=intrinsics['size'][0],
                height=intrinsics['size'][1],
            )
            if camera.calibration is None:
                camera.calibration = Calibration(
                    intrinsics=Intrinsics(
                        matrix=intrinsics['K'],
                        distortion=intrinsics['D'][0],
                        rotation=Rotation(R=intrinsics['R0']),
                        size=size,
                    ),
                )
            if camera.calibration.extrinsics.tilt is None:
                camera.calibration.extrinsics.tilt = tilt
            id = str(uuid.uuid4())
            micros = float(header['micros'])
            if micros < camera.synchronization.micros:
                self.log.warning('Camera needs re-syncronization')
                camera.synchronization = None
                return
            image_time = 1e-6 * micros + camera.synchronization.offset
            world.images.append(Image(id=id, time=image_time, mac=mac, size=size))
            world.image_data[id] = content

        while world.images and world.images[0].time < world.time - 5 * 60.0:
            del world.image_data[world.images[0].id]
            del world.images[0]

    def get(self, url, timeout=1):

        content = io.BytesIO()
        headers = io.BytesIO()

        connection = pycurl.Curl()
        connection.setopt(pycurl.TIMEOUT, timeout)
        connection.setopt(pycurl.URL, url)
        connection.setopt(pycurl.HTTPGET, 1)
        connection.setopt(pycurl.WRITEFUNCTION, content.write)
        connection.setopt(pycurl.HEADERFUNCTION, headers.write)
        try:
            connection.perform()
        finally:
            connection.close()

        content_ = content.getvalue()
        headers_ = dict(line.split(': ', 1) for line in headers.getvalue().decode().splitlines() if ': ' in line)

        return headers_, content_

    def synchronize(self, world: World, camera: Camera, *, attempts: int = 3) -> bool:

        for _ in range(attempts):
            try:
                url = f'http://{camera.network.ip}:8080/micros'
                response = requests.get(url, timeout=(0.5, 0.02))
                now = world.time
                micros = float(response.content.decode())
                if camera.synchronization is None or micros < camera.synchronization.micros:
                    offset = now - 1e-6 * micros
                else:
                    offset = 0.1 * (now - 1e-6 * micros) + 0.9 * camera.synchronization.offset
                return CameraSynchronization(
                    time=now,
                    micros=micros,
                    offset=offset,
                )
            except Exception as e:
                self.log.warning(e)
        self.log.warning("Too many attempts")

    @staticmethod
    def parse_imu(values: str) -> list[float]:

        words = values.split()
        return [float(words[0]), float(words[1]), float(words[2])]

    @staticmethod
    def compute_tilt(imu_calibration: list[list[float]], imu_reading: list[float]) -> Rotation:

        imu_calibration_array = np.array(imu_calibration).T
        calibration_targets = np.array([
            [0, 0, +1],
            [0, 0, -1],
            [0, +1, 0],
            [0, -1, 0],
            [+1, 0, 0],
            [-1, 0, 0],
        ]).T
        m = np.mean(imu_calibration_array, axis=1)[:, None]
        M = np.linalg.lstsq((imu_calibration_array - m).T, calibration_targets.T, rcond=-1)[0].T

        z_dir = np.dot(M, np.array(imu_reading)[:, None] - m).flatten()
        z_dir = z_dir / np.linalg.norm(z_dir)

        x_dir = np.cross([0, 1, 0], z_dir)
        x_dir = x_dir / np.linalg.norm(x_dir)
        y_dir = np.cross(z_dir, x_dir)
        R = np.vstack((x_dir, y_dir, z_dir))
        return Rotation(R=R.tolist())
