import asyncio
import numpy as np
from collections import Counter
import uuid
from ..world.link import Link
from ..world.world import World
from .actor import Actor
from .calibration.helmert_adjustment import HelmertAdjustment
from .calibration.bundle_adjustment import BundleAdjustment


class CameraLinker(Actor):
    interval: float = 0.1
    is_dirty: bool = True
    linked_cameras: list[str] = []
    use_3d_adjustment: bool = False

    async def step(self, world: World):
        if world.link_queue:
            self.collect_new_links(world)
        if any(world.links) and self.is_dirty:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, lambda: self.run_adjustment(world))

    def run_adjustment(self, world: World):
        if self.use_3d_adjustment:
            self.run_bundle_adjustment(world)
        else:
            self.run_helmert_adjustment(world)

    def run_helmert_adjustment(self, world: World):
        ha = HelmertAdjustment()
        ha.load(world)
        if set(ha.frames) != set(self.linked_cameras):
            ha.find_approximation()

        try:
            sigma_0_estimated = ha.adjust()
            self.log.info(f'sigma_0_estimated: {sigma_0_estimated:.3f}')
        except np.linalg.LinAlgError as e:
            self.log.error(e)

        max_dist, max_angle = ha.dump(world)
        self.is_dirty = max_dist > 0.001 or max_angle > np.deg2rad(0.001)
        self.linked_cameras = list(ha.frames.keys())

    def run_bundle_adjustment(self, world: World):
        ha = HelmertAdjustment()
        ha.load(world)
        if set(ha.frames) != set(self.linked_cameras):
            ha.find_approximation()
            ha.dump(world)

        ba = BundleAdjustment()
        ba.load(world)
        sigma_0_estimated = ba.adjust(1)
        self.log.info(f'sigma_0_estimated: {sigma_0_estimated:.3f}')

        max_dist, max_angle = ba.dump(world)
        self.is_dirty = max_dist > 0.001 or max_angle > np.deg2rad(0.001)
        self.linked_cameras = list(ha.frames.keys())

    def collect_new_links(self, world: World):
        while len(world.link_queue) > 0:
            self.is_dirty = True
            set_id = str(uuid.uuid4())
            image_ids = world.link_queue.pop(0)
            for image_id in image_ids:
                for image in world.images:
                    if image.id == image_id:
                        break
                else:
                    continue

                calibration = world.cameras[image.mac].calibration
                if calibration is None or not calibration.is_complete:
                    continue

                for name, image_point in image.marker_points.items():
                    world_point = calibration.project_from_image(image_point, target_height=world.robot.marker.height)
                    if world_point is None:
                        continue
                    link = Link(mac=image.mac, id=f'{set_id}.{name}', image_point=image_point, world_point=world_point)
                    world.links.append(link)

        id_counter = Counter(link.id for link in world.links)
        world.links[:] = [link for link in world.links if id_counter[link.id] > 1]
