import io
import asyncio
from uuid import uuid4
import numpy as np
import PIL as pil
from ..world.image import Image, Size
from ..world.world import World
from ..world.camera import Camera
from ..world.calibration import Calibration, Extrinsics, Intrinsics
from ..world.rotation import Rotation
from .actor import Actor


class CameraSimulator(Actor):

    interval: float = 0.01

    @staticmethod
    def create_random_camera() -> Camera:

        c = 570
        size = Size(width=1600, height=1200)
        K = [[c, 0, size.width / 2], [0, c, size.height / 2], [0, 0, 1]]
        D = [0, 0, 0, 0, 0]
        rotation = Rotation.zero()
        intrinsics = Intrinsics(matrix=K, distortion=D, rotation=rotation, size=size)

        tilt = Rotation.from_euler(0.1 * np.random.randn(), np.pi + 0.1 * np.random.randn(), 0)
        yaw = np.random.randn()
        t = [2 * np.random.randn(), 2 * np.random.randn(), 3 + 0.2 * np.random.randn()]
        extrinsics = Extrinsics(tilt=tilt, yaw=yaw, translation=t)

        return Camera(
            mac=':'.join([f'{np.random.randint(255):02x}' for _ in range(6)]),
            calibration_simulation=Calibration(intrinsics=intrinsics, extrinsics=extrinsics),
        )

    async def step(self, world: World):

        if any(world.download_queue):
            mac = world.download_queue.pop(0)
            camera = world.cameras[mac]

            await asyncio.sleep(1.0)

            if camera.calibration is None:
                camera.calibration = camera.calibration_simulation.copy(deep=True)
                camera.calibration.extrinsics.translation[0] += np.random.randn()
                camera.calibration.extrinsics.translation[1] += np.random.randn()
                camera.calibration.extrinsics.yaw += np.pi * list(world.cameras.values()).index(camera)

            if camera.calibration.extrinsics.tilt is None:
                camera.calibration.extrinsics.tilt = camera.calibration_simulation.extrinsics.tilt

            id = str(uuid4())
            size = camera.calibration.intrinsics.size
            world.images.append(Image(id=id, time=world.time, mac=camera.mac, size=size))

            img = pil.Image.new('RGB', size=(size.width, size.height), color=camera.color)
            img_byte_arr = io.BytesIO()
            img.save(img_byte_arr, format='JPEG')
            world.image_data[id] = img_byte_arr.getvalue()

        while world.images and world.images[0].time < world.time - 5 * 60.0:
            del world.images[0]
