import socketio
import asyncio
from .actor import Actor
from ..world.world import World
from ..world.image import Image
from ..world.detection import Detection
from .. import helpers
from datetime import datetime, timedelta


class Detector(Actor):
    interval: float = 0.02

    def __init__(self):
        super().__init__()
        self.session = None
        self.sio = socketio.AsyncClient()
        self.is_connected = None

        @self.sio.event
        def connect():
            self.is_connected = True
            assert self.sio.transport() == 'websocket'

        @self.sio.event
        def connect_error(data):
            self.is_connected = False

        @self.sio.event
        def disconnect():
            self.is_connected = False

    async def connect(self):
        try:
            self.log.info('connecting to detector')
            await self.sio.connect("ws://localhost:8004", socketio_path="/ws/socket.io")
            self.log.info('connected successfully')
            self.is_connected = True
        except:
            self.log.exception('connection failed; trying again')
            self.is_connected = None
            await asyncio.sleep(3)

    async def tear_down(self):
        await self.sio.disconnect()

    async def step(self, world: World):
        if self.is_connected is None:
            await self.connect()
        if self.is_connected is True:
            for image in world.upload.queue.values():
                if datetime.now() < world.upload.last_upload + timedelta(minutes=world.upload.minimal_minutes_between_uploads):
                    break
                if image.id in world.image_data:
                    world.upload.last_upload = datetime.now()
                    world.upload.queue.clear()  # because old images should not be uploaded later when the robot is inactive
                    await self.upload(image, world.image_data[image.id])
                    break
            for image in reversed(world.images):
                if image.detections is None:
                    await self.detect(image, world.image_data[image.id])
                    break

    async def detect(self, image: Image, image_data: bytes):
        try:
            result = await self.sio.call('detect', {'image': image_data, 'mac': image.mac})
            if 'box_detections' in result:
                image.detections = [Detection.parse_obj(d) for d in result['box_detections']]
            else:
                self.log.error(result)
        except:
            self.log.exception(f'could not detect for {image.mac}')

    async def upload(self, image: Image, image_data: bytes):
        try:
            await self.sio.emit('upload', {'image': image_data, 'mac': image.mac})
        except:
            self.log.exception(f'could not upload for {image.mac}')

    def __str__(self) -> str:
        state = {
            None: 'starting',
            True: 'connected',
            False: 'reconnecting',
        }[self.is_connected]
        return f'{type(self).__name__} ({state})'
