import numpy as np
from .actor import Actor
from ..world.point import Point
from ..world.pose import Pose
from ..world.world import World


class RobotLocator(Actor):
    last_image = None

    def find_robot(self, world: World):
        if not any(world.images):
            return
        image = world.images[-1]
        if image == self.last_image:
            return
        self.last_image = image
        camera = world.cameras.get(image.mac)
        if camera is None:
            return
        calibration = camera.calibration
        if calibration is None:
            return
        if image.detections is None:
            return

        expected_points = {key: Point(x=value[0], y=value[1]) for key, value in world.robot.marker.points.items()}
        detected_image_points = {}
        detected_world_points = {}
        for category in expected_points.keys():
            detections = [d for d in image.detections if d.category_name == category and d.confidence >= 0.3]
            if not any(detections):
                continue
            if len(detections) > 1:
                msg = f'adding {image.id} to upload queue because it has multiple detections of the same category:\n   '
                self.log.warning(msg + '\n   '.join([str(d) for d in detections]))
                world.upload.mark(image)
            detection = sorted(detections, key=lambda d: d.confidence)[-1]
            image_point = Point(x=detection.cx, y=detection.cy)
            world_point = calibration.project_from_image(image_point, world.robot.marker.height)
            if world_point is None:
                continue
            detected_image_points[category] = image_point
            detected_world_points[category] = Point(x=world_point.x, y=world_point.y)

        if 0 < len(detected_world_points) < len(expected_points):
            self.log.warning(f'adding {image.id} to upload queue because it has to few detections')
            world.upload.mark(image)
        if len(detected_world_points) < 2:
            self.log.warning(f'Too few points: {len(detected_world_points)}')
            image.marker_points = {}
            return

        err_limit = 0.01
        X, Y, Yaw, err = self.motion_from_point_pairs(expected_points, detected_world_points)
        plausible = err < err_limit
        if not plausible and \
            "marker_hinten_links" in detected_world_points and \
                "marker_hinten_rechts" in detected_world_points:
            detected_world_points["marker_hinten_rechts"], detected_world_points["marker_hinten_links"] = \
                detected_world_points["marker_hinten_links"], detected_world_points["marker_hinten_rechts"]
            detected_image_points["marker_hinten_rechts"], detected_image_points["marker_hinten_links"] = \
                detected_image_points["marker_hinten_links"], detected_image_points["marker_hinten_rechts"]
            X, Y, Yaw, err = self.motion_from_point_pairs(expected_points, detected_world_points)
            plausible = err < err_limit
        if not plausible:
            self.log.warning(f'Error too large: {err}')
            image.marker_points = {}
            self.log.warning(f'adding {image.id} to upload queue because detections are not plausible')
            world.upload.mark(image)
            return

        world.robot.detection = Pose(x=X, y=Y, yaw=Yaw, time=image.time)
        image.marker_points = detected_image_points

    def motion_from_point_pairs(self, expected_points: dict[str, Point], detected_points: dict[str, Point]) -> tuple:
        # http://nghiaho.com/?page_id=671
        A = np.array([expected_points[category].tuple for category in expected_points.keys()])
        B = np.array([detected_points[category].tuple for category in detected_points.keys()])
        center_A = np.mean(A, axis=0)
        center_B = np.mean(B, axis=0)

        H = (A - center_A).T.dot(B - center_B)
        U, _, VT = np.linalg.svd(H)
        D = np.diag((1, 1 if np.linalg.det(U.dot(VT)) > 0 else -1))
        R = VT.T.dot(D).dot(U.T)
        Yaw = np.arctan2(R[1, 0], R[0, 0])
        t = center_B - center_A.dot(R.T)
        err = np.linalg.norm(np.dot(A, R.T) + t - B) / (A.size - 1)

        return t[0], t[1], Yaw, err
