from nicegui.elements.button import Button
from nicegui.elements.card import Card
from nicegui.elements.checkbox import Checkbox
from nicegui.elements.dialog import Dialog
from nicegui.elements.html import Html
from nicegui.elements.icon import Icon
from nicegui.elements.image import Image
from nicegui.elements.label import Label
from nicegui.elements.number import Number
from nicegui.elements.row import Row
from nicegui.elements.svg import Svg
from nicegui.elements.switch import Switch
from nicegui.timer import Timer
from ..runtime import Runtime
from ..world.camera import Camera
from ..actors.camera_simulator import CameraSimulator


class CameraCard(Card):

    def __init__(self, runtime: Runtime):

        super().__init__()

        self.runtime = runtime
        self.world = runtime.world
        self.linkable_images = {}
        self.edit_camera = None

        with self:

            with Dialog() as self.edit_dialog, Card():

                with Row().classes('items-end'):
                    self.height_input = Number('Height [cm]')
                    Button(on_click=self.close_edit_dialog).props('icon=done outline round size=sm')
                Button('Delete calibration', on_click=self.delete_calibration).props('outline color=red')
                Button('Delete links', on_click=self.delete_links).props('outline color=red')
                Button('Delete images', on_click=self.delete_images).props('outline color=red')
                Button('Delete camera', on_click=self.delete_camera).props('outline color=red')

            with Dialog().props('auto-close') as self.clear_dialog, Card().classes('items-stretch'):

                Button('Clear calibrations', on_click=self.delete_all_calibrations).props('outline color=red')
                Button('Clear links', on_click=self.delete_all_links).props('outline color=red')
                Button('Clear images', on_click=self.delete_all_images).props('outline color=red')
                Button('Clear cameras', on_click=self.delete_all_cameras).props('outline color=red')

            with Dialog().props('maximized auto-close') as self.image_dialog, Card():

                pass

            with Row().classes('items-center'):

                Button('Add camera', on_click=self.add_camera).props('icon=add flat')
                Button('Download', on_click=lambda: self.download(list(self.world.cameras.keys()))) \
                    .props('icon=file_download flat')
                self.track_check = Checkbox('Track', on_change=lambda e: self.track_all(e.value))
                self.auto_switch = Switch('Auto track', on_change=lambda e: self.auto_track(e.value))
                Checkbox('3D adjustment').bind_value(runtime.camera_linker.use_3d_adjustment)
                self.link_button = Button('Link', on_click=self.link).props('icon=link flat rounded')
                Button(on_click=self.clear_dialog.open).props('icon=more_vert flat round')

            self.camera_grid = Row()

            Timer(1.0, self.update_cameras)

    def add_camera(self, _):

        camera = CameraSimulator.create_random_camera()
        self.world.cameras[camera.mac] = camera

    def delete_all_cameras(self):

        mac = self.edit_camera.mac
        del self.world.cameras[mac]
        for image in self.world.images:
            if image.mac == mac:
                del self.world.images[self.world.images.index(image)]
                del self.world.image_data[image.id]
        for link in self.world.links:
            if link.mac == mac:
                del self.world.links[self.world.links.index(link)]
        self.world.link_queue.clear()
        self.edit_dialog.close()

    def delete_all_calibrations(self):

        for camera in self.world.cameras.values():
            camera.calibration = None
            camera.projection = None
        self.delete_all_images()
        self.delete_all_links()

    def delete_all_links(self):

        self.world.links.clear()
        self.world.link_queue.clear()

    def delete_all_cameras(self):

        self.world.cameras.clear()
        self.delete_all_images()
        self.delete_all_links()

    def delete_all_images(self):

        self.world.images.clear()
        self.world.image_data.clear()

    def delete_calibration(self):

        self.edit_camera.calibration = None
        self.edit_camera.projection = None
        self.delete_images()
        self.delete_links()
        self.edit_dialog.close()

    def delete_links(self):

        ids = [link.id.split('_')[0] for link in self.world.links if link.mac == self.edit_camera.mac]
        self.world.links[:] = [link for link in self.world.links if link.id.split('_')[0] not in ids]
        self.edit_dialog.close()

    def delete_camera(self):

        del self.world.cameras[self.edit_camera.mac]
        self.delete_images()
        self.delete_links()
        self.edit_dialog.close()

    def delete_images(self):

        ids = [image.id for image in self.world.images if image.mac == self.edit_camera.mac]
        self.world.images[:] = [image for image in self.world.images if image.id not in ids]
        self.world.image_data = {id: data for id, data in self.world.image_data.items() if id not in ids}
        self.edit_dialog.close()

    def track_all(self, activate: bool):

        self.world.tracking = list(self.world.cameras) if activate else []
        self.auto_switch.value = False

    def auto_track(self, activate: bool):

        self.world.tracking = activate
        if activate:
            self.track_check.value = False

    def link(self):

        self.world.link_queue.append(list(self.linkable_images.values()))

    def download(self, macs: list[str]):

        if self.world.tracking:
            self.world.tracking = False
            self.auto_switch.value = False
        self.world.download_queue.extend(mac for mac in macs if mac not in self.world.download_queue)
        return False

    def track(self, activate: bool, mac: str):

        if not isinstance(self.world.tracking, list):
            self.world.tracking = []
            self.auto_switch.value = False

        if activate and mac not in self.world.tracking:
            self.world.tracking.append(mac)

        if not activate and mac in self.world.tracking:
            del self.world.tracking[self.world.tracking.index(mac)]

    def open_edit_dialog(self, camera: Camera):

        self.edit_camera = camera
        if camera.calibration is not None:
            self.height_input.value = round(camera.calibration.extrinsics.translation[2] * 100)
            self.height_input.visible = True
        else:
            self.height_input.visible = False
        self.edit_dialog.open()

    def close_edit_dialog(self):

        if self.edit_camera.calibration is not None:
            self.edit_camera.calibration.extrinsics.translation[2] = self.height_input.value / 100
            self.edit_camera.calibration.extrinsics.tilt = None
        self.edit_dialog.close()

    def open_image_dialog(self, image: Image):

        self.image_dialog.view.delete_components()
        with self.image_dialog:
            self.insert_image(image)
        self.image_dialog.open()

    def toggle_lock(self, mac: str):

        self.world.cameras[mac].locked = not self.world.cameras[mac].locked
        self.runtime.camera_linker.is_dirty = True

    def insert_image(self, image: Image):

        with Image(f'imagedata/{image.id}').classes('w-full').props('contain') as image_element:
            if image.detections is not None:
                svg_content = '<svg viewBox="0 0 1600 1200" width="100%" height="100%" xmlns="http://www.w3.org/2000/svg">'
                for d in image.detections:
                    letter = d.category_name.replace('_', ' ').split()[-1][0].upper()
                    svg_content += f'<rect x="{d.x}" y="{d.y}" width="{d.width}" height="{d.height}" stroke="red" fill="red" fill-opacity="10%" />'
                    svg_content += f'<text x="{d.x+2}" y="{d.y+d.height-2}" fill="white">{letter}</text>'
                svg_content += '</svg>'
                Svg(svg_content).style('background:transparent')
        return image_element

    def update_cameras(self):

        self.linkable_images.clear()

        self.camera_grid.view.delete_components()
        with self.camera_grid:

            for mac, camera in self.world.cameras.items():
                with Card().style('width:20em;height:23.5em'):

                    with Row().classes('w-full items-center justify-between'):
                        Html(f'<a href="http://{camera.network.ip}:8080" target="_blank">{mac}</a>')
                        if camera.calibration is not None and camera.calibration.extrinsics is not None:
                            Label(f'{round(camera.calibration.extrinsics.translation[2] * 100)} cm')
                        Button(on_click=lambda _, camera=camera: self.open_edit_dialog(camera)) \
                            .props('icon=edit flat round size=sm')

                    with Row().classes('w-full items-center') as camera_row:
                        Button(on_click=lambda mac=mac: self.download([mac])).props('icon=file_download flat round')
                        Checkbox(value=isinstance(self.world.tracking, list) and mac in self.world.tracking,
                                 on_change=lambda e, mac=mac: self.track(e.value, mac))
                        Button(on_click=lambda _, mac=mac: self.toggle_lock(mac)) \
                            .props(f'flat round icon={"lock" if camera.locked else "lock_open"}')

                    images = [i for i in self.world.images if i.mac == mac and i.id in self.world.image_data]
                    if not any(images):
                        continue
                    image = images[-1]
                    self.insert_image(image)

                    linkable = [i.id for i in images if i.marker_points and i.time > self.world.robot.last_movement]
                    if any(linkable):
                        self.linkable_images[mac] = linkable[-1]
                    with camera_row:
                        Icon('link').classes('text-green-600' if mac in self.linkable_images else 'text-red-600')
                        Button(on_click=lambda _, image=image: self.open_image_dialog(image)) \
                            .props('flat round icon=open_in_new')

        num = len(self.linkable_images)
        self.link_button.set_text(f'Link ({num})' if num else 'Link')
        self.link_button.view.disable = num < 2
