import numpy as np
from nicegui.elements.scene_object3d import Object3D
from nicegui.elements.scene_objects import Cylinder, Group, Sphere, Texture
from rosys.world.camera import Camera
from rosys.world.image import Image
from rosys.world.link import Link


class CameraObject(Object3D):

    def __init__(self, cameras: dict[str, Camera], links: list[Link], images: list[Image]):
        super().__init__('group')
        self.cameras = cameras
        self.links = links
        self.images = images
        self.update()

    @property
    def calibrated_cameras(self):
        return {
            mac: camera
            for mac, camera in self.cameras.items()
            if camera.calibration is not None and camera.calibration.is_complete
        }

    def find_objects(self, type: str) -> dict[str, Object3D]:
        return {
            (obj.name or '').split('_', 1)[1]: obj
            for obj in self.view.objects.values()
            if (obj.name or '').split('_', 1)[0] == type
        }

    def update(self):
        self.update_cameras()
        self.update_links()
        self.update_images()

    def update_cameras(self):
        camera_groups = self.find_objects('camera')
        for mac, camera_group in camera_groups.items():
            if mac not in self.calibrated_cameras:
                camera_group.delete()
        for mac, camera in self.calibrated_cameras.items():
            if mac not in camera_groups:
                with Group().with_name(f'camera_{mac}') as camera_groups[mac]:
                    with Group() as pyramid:
                        Cylinder(0, np.sqrt(0.5), 1, 4) \
                            .rotate(-np.pi / 2, 0, np.pi / 4) \
                            .move(z=0.5)
                    pyramid.scale(
                        0.001 * camera.calibration.intrinsics.size.width,
                        0.001 * camera.calibration.intrinsics.size.height,
                        0.001 * camera.calibration.intrinsics.matrix[0][0],
                    )
            camera_groups[mac].move(*camera.calibration.extrinsics.translation)
            camera_groups[mac].rotate_R(camera.calibration.rotation.R)

    def update_links(self):
        links = {
            f'{link.mac}.{link.id}': link
            for link in self.links
        }
        spheres = self.find_objects('link')
        for id, sphere in spheres.items():
            if id not in links:
                sphere.delete()
        for id, link in links.items():
            if id not in spheres:
                spheres[id] = Sphere(0.05).with_name(f'link_{id}').material('#ff0000')
            spheres[id].move(link.world_point.x, link.world_point.y, link.world_point.z)

    def update_images(self):
        images = {image.mac: image for image in self.images}
        image_ids = [image.id for image in images.values()]
        textures = self.find_objects('image')
        for id, texture in textures.items():
            if id not in image_ids:
                texture.delete()
        for image in images.values():
            if image.id not in textures:
                camera = self.calibrated_cameras.get(image.mac)
                if camera is not None and camera.projection is not None:
                    url = f'imagedata/{image.id}'
                    coordinates = [[point and [point[0], point[1], 0] for point in row] for row in camera.projection]
                    textures[image.id] = Texture(url, coordinates).with_name(f'image_{image.id}')
